""" Database connection """

# Original
# __version__ = "Revision: 1.23 "
# Source: /home/vladap/src/mysqlUserFolder/RCS/db.py,v 
#
# Author: vladap 
# Date: 2003/01/20 21:03:23 

import string, re
import util, cfg
from Products.PythonScripts.standard import sql_quote

# --------------------------------------------------------------------- */
# Utility functions


def setup_user_roles (roles):
	roles.append (cfg.DBUSERFOLDER_ROLE)
	roles.append (cfg.ZOPE_ANONYMOUS_ROLE)
	roles.append (cfg.ZOPE_AUTHENTICATED_ROLE)
			
	iroles = cfg.VALIDATE_IGNORE_ROLES
	
	if len (iroles) > 0:
		new_roles = filter (lambda x, iroles=iroles: not (x in iroles), roles)
	else:
		new_roles = roles

	return new_roles


def split_params (str_param):
    r = {}
    s = string.strip(str_param)
    for token in re.split('\s+', s):
        (key, value) = string.split(token, '=')
        r[key] = value
    return r

# --------------------------------------------------------------------- */
# dbConnection class

class dbConnection:
    """ DB Connection class. 
    
        It is supposed to be refferenced as _v_ attribute of all classes
        that need access to a database. So this class does not use locks.
    """  

    def __init__ (Self, DSN, dbtype):
        dbtype_dic = { # dbtype_module, dbtype_sql, dbtype_thread
            'psycopg' : ('psycopg', 'PostgreSQL', 0),
            'DCOracle2' : ('DCOracle2', 'oracle', 1),
            'DCOracle2-MT' : ('DCOracle2', 'oracle', 2),
            'cx_Oracle' : ('cx_Oracle', 'oracle', 1),
            'cx_Oracle-MT' : ('cx_Oracle', 'oracle', 2),
            'kinterbasdb' : ('kinterbasdb', 'Firebird', 1),
            'kinterbasdb-MT' : ('kinterbasdb', 'Firebird', 2),
            'sqlite' : ('sqlite', 'SQLite', 1),
            'sqlite-MT' : ('sqlite', 'SQLite', 2),
        }
        Self.DSN = DSN
        Self.dbtype = dbtype
        (Self.dbtype_module, Self.dbtype_sql, Self.dbtype_thread) = \
                                            dbtype_dic[Self.dbtype]
        if Self.dbtype_thread == 1:
            import thread
            Self.lock = thread.allocate_lock()

        Self.db_con = None
    #--------------------------------------------
    # Internal general utility functions
    def __query_list_field (Self, cur, query):
        """ This function returns a list of values from the select query.
            Result set should contain only one field. """
        Self.execute (cur, query)
        data = cur.fetchall ()
        ret = []
        
        for i in range (len (data)):
            (value,) = data [i]
            ret.append (value)
        
        return ret

    def __get_nextval(Self, cur, seq_name):
        ret = 0
        if Self.dbtype_sql == 'PostgreSQL':
            query = "select nextval('%s')" % (sql_quote(seq_name), )
            Self.execute (cur, query)
            ret = cur.fetchone()[0]
        elif Self.dbtype_sql == 'oracle':
            query = "select %s.nextval from dual" % (seq_name,)
            Self.execute (cur, query)
            ret = cur.fetchone()[0]
        elif Self.dbtype_sql == 'Firebird':
            query = "select gen_id(%s,1) from rdb$database" %(seq_name,)
            Self.execute (cur, query)
            ret = cur.fetchone()[0]
        else:
            query = "select NextVal from ZSeq where Category='%s'" %(seq_name, )
            Self.execute (cur, query)
            ret = cur.fetchone()[0]
            Self.execute(cur, "update ZSeq set NextVal=NextVal+1 where Category='%s'" %(seq_name, ))
        return int(ret)

        
    #--------------------------------------------
    # connect / dicconnect 
    
    def connect (Self):
        if Self.dbtype_thread == 0:
            exec 'import ' + Self.dbtype_module + ' as dbmodule'
            Self.db_con = dbmodule.connect (Self.DSN, serialize=0)
            util.log_debug ("db.connect (): " + Self.DSN)
        elif Self.dbtype_thread == 1:
            exec 'import ' + Self.dbtype_module + ' as dbmodule'
            if Self.dbtype_module == 'kinterbasdb':
                Self.db_con = dbmodule.connect (**split_params(Self.DSN))
                tpb = dbmodule.isc_tpb_write + dbmodule.isc_tpb_concurrency
                Self.db_con.begin(tpb)
            else:
                Self.db_con = dbmodule.connect (Self.DSN)
            util.log_debug ("db.connect (): " + Self.DSN)
        elif Self.dbtype_thread == 2:
            pass
        return 1
    
    def disconnect (Self):
        if not Self.db_con is None:
            Self.db_con.close ()
            
        Self.db_con = None
    
    #--------------------------------------------
    # cursor / execute / commit / rollback
    def cursor (Self):
        if Self.dbtype_thread == 1:
            Self.lock.acquire ()
        elif Self.dbtype_thread == 2:
            exec 'import ' + Self.dbtype_module + ' as dbmodule'

            if Self.dbtype_module == 'kinterbasdb':
                Self.db_con = dbmodule.connect (**split_params(Self.DSN))
                tpb = dbmodule.isc_tpb_write + dbmodule.isc_tpb_concurrency
                Self.db_con.begin(tpb)
            else:
                Self.db_con = dbmodule.connect (Self.DSN)
            util.log_debug ("db.connect ():" + Self.DSN)
        elif Self.db_con is None:
            Self.connect()
        return Self.db_con.cursor ()

    def execute (Self, cur, query):
        util.log_debug ("execute():" + query)
        cur.execute (query)

    def commit (Self, cur):
        if Self.dbtype_thread == 0:
            cur.commit ()
        else:
            Self.db_con.commit ()
        if Self.dbtype_thread == 1:
            Self.lock.release ()
        elif Self.dbtype_thread == 2:
            if not Self.db_con is None:
                Self.db_con.close ()
            Self.db_con = None
    
    def rollback (Self, cur):
        if Self.dbtype_thread == 0:
            cur.rollback ()
        else:
            Self.db_con.rollback ()
        if Self.dbtype_thread == 1:
            Self._release ()
        elif Self.dbtype_thread == 2:
            if not Self.db_con is None:
                Self.db_con.close ()
            Self.db_con = None
    
    #--------------------------------------------
    # Internal User db functions.
    
    def __check_user (Self, user):
        username = user.getUserName ()
        realm = user.getRealm ()
        id = user.getDBId ()
        query = "select count(*) from ZUsers " \
            "where Username = '%s' and Realm = '%s' and Id = '%i'" % \
            (sql_quote(username), sql_quote(realm), id)
                                                
        cur = Self.cursor ()
        Self.execute (cur, query)
        cnt = cur.fetchone ()[0]
        Self.commit (cur)
        
        return (cnt == 1)
        
    def __check_user_exist (Self, username, realm):
        query = "select count(*) from ZUsers " \
            "where Username = '%s' and Realm = '%s'" % \
            (sql_quote(username), sql_quote(realm))
        cur = Self.cursor ()
        Self.execute (cur, query)

        cnt = cur.fetchone ()[0]
        Self.commit (cur)
        
        return (cnt == 1)
        
    #--------------------------------------------
    # User db API functions.
        
    def list_rolenames (Self, realm):
        util.log_debug ("list_rolenames()")
        cur = Self.cursor ()
        query = "select Rolename from ZRoles " \
            "where Realm = '%s' order by Rolename" % (sql_quote(realm), )
            
        ret = Self.__query_list_field (cur, query)
        Self.commit (cur)
        
        return ret
        
    def list_usernames (Self, realm):
        util.log_debug ("list_usernames()")
        cur = Self.cursor ()
        query = "select Username from ZUsers " \
            "where Realm = '%s' order by Username" % (sql_quote(realm), )
        ret = Self.__query_list_field (cur, query)
        Self.commit (cur)
        
        return ret


    def list_usernames_with_role (Self, realm, role):
        util.log_debug ("list_rolenames_with_role()")
        cur = Self.cursor ()
        query = "select id from ZRoles " \
            "where Realm = '%s' and Rolename = '%s'" % (realm, role)

        [rid, ] = Self.__query_list_field (cur, query)

        query = "Select ZUsers.Username from ZUsers,ZUserRoles where " \
            "(ZUsers.Id = ZUserRoles.Id_User) and (ZUserRoles.Id_Role = '%i')" \
            "and (ZUsers.Realm = '%s') order by Username" % \
            (rid, sql_quote(realm))
        ret = Self.__query_list_field (cur, query)
        Self.commit (cur)
        return ret
        
    def get_user_data (Self, user, by_id = 0):
        util.log_debug ("get_user_data()")
        username = user.getUserName ()
        realm = user.getRealm ()
        
        if by_id: uid = user.getDBId ()
        
        user._setDBId (cfg.VDB_INVALID_ID)
        user._setPassword ('', 0)
        user._setAuthData ([], [])
        user._setOtherData ('<unknown>', '<unknown>')
    
            # ----- Trazenje usera
        cur = Self.cursor ()
            
        if by_id:
            query = "select Username, Realm from ZUsers " \
                "where Id = '%i'" % (uid, )
            Self.execute (cur, query)
            data = cur.fetchall ()
            if len (data) == 0: 
                util.log_debug ("dbConn.get_user_data (): " 
                                            "no user data by id: '%i'" % uid)
                Self.commit (cur)
                return 0
                
            (du, dr) = data [0]
            username = du
            user._setUserName (username)
            
            if dr != realm:
                util.log_debug ("dbConn.get_user_data (): "
                                                "invalid realm for user id")
                return 0
            
        query = "select Id, EncPassword, PasswordType, EMail, Realname, " \
                "Domains from ZUsers where Realm = '%s' and Username = '%s'" % \
                    (sql_quote(realm), sql_quote(username))
        
        Self.execute (cur, query)
        data = cur.fetchall ()
        if len (data) == 0: 
            util.log_debug ("dbConn.get_user_data (): no user data")
            Self.commit (cur)
            return 0
                  
        (id, password, password_type, email, realname, domains_str) = data [0]

        if domains_str:
            domains = string.split (string.strip (domains_str))
        else:
            domains = []
        
            # ----- Find roles 
        
        query = "Select Id_Role from ZUserRoles where Id_User = %i" % (id, )
        role_ids = Self.__query_list_field (cur, query)
    
        roles = []
        
        for role_id in role_ids:
            query = "Select Rolename from ZRoles where Id = %i" % (role_id, )
            (role,) = Self.__query_list_field (cur, query)
            
            roles.append (role)
            
        roles = setup_user_roles (roles)
        
            # ----- Popunjavanje strukture

            
        user._setDBId (id)
        user._setOtherData (realname, email)
        user._setAuthData (roles, domains)
        user._setPassword (password, password_type)
        
        Self.commit (cur)
        return 1
        
    def count_users (Self, realm):
        util.log_debug ("count_users()")
        query = "select count(*) from ZUsers where Realm = '%s'" % \
            (sql_quote(realm), )
        cur = Self.cursor ()
        Self.execute (cur, query)
        cnt = int(cur.fetchone ()[0])
        Self.commit (cur)
        return cnt
        
    def count_roles (Self, realm):
        util.log_debug ("count_roles()")
        query = "select count(*) from ZRoles where Realm = '%s'" % \
            (sql_quote(realm), )
        cur = Self.cursor ()
        Self.execute (cur, query)
        cnt = int(cur.fetchone ()[0])
        Self.commit (cur)
        return cnt
        
    def change_user_password (Self, user, password, ptype):
        util.log_debug ("change_user_password()")
        if not Self.__check_user (user):
            return 0
    
        id = user.getDBId ()
    
        cur = Self.cursor ()
        query = "update ZUsers set " \
            "EncPassword = '%s', PasswordType = '%i' where Id = '%i'" % \
            (sql_quote(password), ptype, id)
                                
        Self.execute (cur, query)
        ret = (cur.rowcount == 1)
            
        Self.commit(cur)
        return ret

    def change_user_roles (Self, user, roles):
        util.log_debug ("change_user_roles()")
        if not Self.__check_user (user):
            return 0

        user_id = user.getDBId ()
        realm = user.getRealm ()
        
        cur = Self.cursor ()
        query = "delete from ZUserRoles where Id_User = %i" % (user_id, )
        Self.execute (cur, query)

        for role in roles:
            query = "select Id from ZRoles " \
                "where Rolename = '%s' and Realm = '%s'" % \
                (sql_quote(role), sql_quote(realm))
            ret = Self.__query_list_field (cur, query)
            if ret != []:
                [role_id] = ret
            
                query = "insert into ZUserRoles (Id_Role, Id_User) " \
                    "Values ('%i', '%i')" % \
                    (role_id, user_id)
                Self.execute (cur, query)
            else:
                Self.rollback (cur)
                raise 'Unknown role: %s' % (query)

        Self.commit (cur)
        return 1

    def change_user_domains (Self, user, domains):
        util.log_debug ("change_user_domains()")
        if type (domains) == type ([]):
            domains = string.join (domains)
            
        if not Self.__check_user (user):
            return 0
    
        id = user.getDBId ()

        cur = Self.cursor()
        query = "update ZUsers set Domains = '%s' where Id = '%i'" % \
            (sql_quote(domains), id)
        Self.execute (cur, query)
        ret = (cur.rowcount == 1)
        
        Self.commit(cur)
        return ret
        
    def change_user_other (Self, user, realname, email):
        util.log_debug ("change_user_other()")
        if not Self.__check_user (user):
            return 0
    
        id = user.getDBId ()
    
        cur = Self.cursor()
        query = "update ZUsers set Realname = '%s', " \
            "EMail = '%s' where Id = '%i'" % \
            (sql_quote(realname), sql_quote(email), id)
        Self.execute (cur, query)
        ret = (cur.rowcount == 1)
        
        Self.commit(cur)
        return ret
        
    def create_user (Self, username, realm, roles):
        util.log_debug ("create_user()")
        if Self.__check_user_exist (username, realm):
            return 0

        cur = Self.cursor()
        nextval = Self.__get_nextval(cur, 'ZUsers_seq')
        query = "insert into ZUsers (Id, Username, Realm, EncPassword, " \
            "PasswordType) values (%i, '%s', '%s', '%s', %i)" % \
            (nextval, sql_quote(username), sql_quote(realm), '', \
            cfg.VDB_INVALID_PASSWORD_TYPE)

        Self.execute (cur, query)
        user_id = nextval

        for role in roles:
            query = "select Id from ZRoles " \
                "where Rolename = '%s' and Realm = '%s'" % \
                (sql_quote(role), sql_quote(realm))

            ret = Self.__query_list_field (cur, query)
            if ret != []:
                [role_id] = ret
            
                query = "insert into ZUserRoles (Id_Role, Id_User)" \
                    "Values ('%i', '%i')" % \
                    (role_id, user_id)
                Self.execute (cur, query)
            else:
                Self.del_user (username, realm)
                Self.commit (cur)
                raise 'Unknown role: (%s, %s)' % (role, realm)
        
        Self.commit(cur)
        return 1
        
    def del_user (Self, username, realm, mclass = None):
        util.log_debug ("del_user()")
        cur = Self.cursor ()
        query = "select Id from ZUsers " \
            "where Username = '%s' and Realm = '%s'" % \
            (sql_quote(username), sql_quote(realm))
                    
        ret = Self.__query_list_field (cur, query)
        if ret == []:
            return 0
            
        [id] = ret

        query = "select Id from ZSessions where Id_User = %i" % (id, )
        sids = Self.__query_list_field (cur, query)
        
        for sid in sids:
            Self.del_token ('session', sid, realm)
        
        Self.del_token ('user', id, realm)
        
        query = "delete from ZSessions where Id_User = %i" % (id, )
        Self.execute (cur, query)
        
        query = "delete from ZUserRoles where Id_User = %i" % (id, )
        Self.execute (cur, query)
        
        query = "delete from ZUsers where Id = %i" % (id, )
        Self.execute (cur, query)
        Self.commit(cur)
        
        if mclass: Self.del_misc_data_class (mclass, id)
        
        return 1

    def create_role (Self, role, realm):
        util.log_debug ("create_role()")
        cur = Self.cursor ()
        nextval = Self.__get_nextval(cur, 'ZRoles_seq')
        query = "insert into ZRoles (Id, Rolename, Realm) " \
            "values (%i, '%s', '%s')" % \
            (nextval, sql_quote(role), sql_quote(realm))
        Self.execute (cur, query)
        Self.commit(cur)
        return 1

    def del_role (Self, role, realm):
        util.log_debug ("del_role()")
        cur = Self.cursor ()
        query = "select Id from ZRoles " \
            "where Rolename = '%s' and Realm = '%s'" % \
            (sql_quote(role), sql_quote(realm))
        ret = Self.__query_list_field (cur, query)

        if ret == []:
            Self.commit (cur)
            return 0

        [role_id] = ret
            
        query = "delete from ZRoles where Id = %i" % (role_id, )
        Self.execute (cur, query)

        query = "delete from ZUserRoles where Id_Role = %i" % (role_id, )
        Self.execute (cur, query)
        Self.commit(cur)
        
        return 1

        
    #--------------------------------------------
    # Session db API functions
        
    def create_session (Self, session):
        util.log_debug ("create_session()")
        user_id = session.getUserId ()
        
        cur = Self.cursor ()
        nextval = Self.__get_nextval(cur, 'ZSessions_seq')
        query = "insert into ZSessions (Id, id_user) " \
            "values (%i, '%s')" % (nextval, user_id)
        Self.execute (cur, query)
        Self.commit(cur)
        
        session._setId (nextval)
        
    def get_session (Self, sess):
        util.log_debug ("get_session()")
        id = sess.getId ()
        
        cur = Self.cursor ()
        query = "select Id_User from ZSessions where Id = %i" % (id, )
        res = Self.__query_list_field (cur, query)
        Self.commit (cur)
        
        if not res:
            sess._setId (cfg.VDB_INVALID_ID)
            sess._setUserId (cfg.VDB_INVALID_ID)
            
            return sess
            
        assert (len (res) == 1)
        sess._setUserId (int(res [0]))
        
    def session_set_user_id (Self, session, user_id):
        util.log_debug ("session_set_user_id()")
        cur = Self.cursor ()
        query = "update ZSessions set Id_User = '%i' where Id = '%i'" % \
            (user_id, session.getId())
        
        Self.execute (cur, query)
        
        Self.commit(cur)
    
    #--------------------------------------------
    # Token db API functions
    
    
    def del_token (Self, class_id, token_id, realm):
        util.log_debug ("del_token()")
        cur = Self.cursor ()
        query = "delete from ZTokens where Class = '%s' " \
            "and Id = '%i' and Realm = '%s'" % \
            (class_id, token_id, realm)
        
        Self.execute (cur, query)

        Self.commit(cur)
        
    def create_token (Self, class_id, token_id, token_value, 
            token_timeout, token_life, created_time, accessed_time, realm):
        util.log_debug ("create_token()")
        Self.del_token (class_id, token_id, realm)
        
        cur = Self.cursor ()
        query = "insert into ZTokens (Class, Id, TokenValue, Created, " \
            "Accessed, Timeout, Life, Realm) values " \
            "('%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s')" % \
            (class_id, token_id, token_value, created_time,
                            accessed_time, token_timeout, token_life, realm)
        
        Self.execute (cur, query)
        
        Self.commit (cur)
        
    def get_token (Self, class_id, token_id, realm):
        util.log_debug ("get_token()")
        cur = Self.cursor ()
        query = "select TokenValue, Created, Accessed, Timeout, Life from ZTokens " \
            "where Class = '%s' and Id = '%s' and Realm = '%s'" % \
            (class_id, token_id, realm)
                        
        Self.execute (cur, query)
        data = cur.fetchall ()
        Self.commit (cur)
        
        if len (data) == 0: return None
        assert (len (data) == 1)
        # Numeric value get as String in pysqlite, so convert them to float.
        ret = [data[0][0], float(data[0][1]), float(data[0][2]), data[0][3], data[0][4]]
        return ret
        
    def update_token_accessed (Self, class_id, token_id, time, realm):
        util.log_debug ("update_token_accessed()")
        cur = Self.cursor ()
        query = "update ZTokens set Accessed = '%s' where Class = '%s' " \
            "and Id = '%s' and Realm = '%s'" % \
            (time, class_id, token_id, realm)
        
        Self.execute (cur, query)

        Self.commit(cur)
        
    #--------------------------------------------
    # MiscData API functions
    
    def save_misc_data (Self, obj_class, obj_id, misc_id, misc_tuple):
        util.log_debug ("save_misc_data()")
        if type (misc_tuple) != type (()):
            raise 'API Error', \
                "save_misc_data () should be called with tuple. "
    
        (iv, sv) = misc_tuple
        
        cur = Self.cursor ()
        data = Self.get_misc_data (obj_class, obj_id, misc_id, check=1)
        if data:
            query = "update ZMiscData set " \
                    "Value_int = %i, Value_string = '%s' where " \
                    "ObjId = '%s' and OClass = '%s' and MId = '%s'" % \
                    (iv, sv, obj_id, obj_class, misc_id)
            
            Self.execute (cur, query)
        else:
            query = "insert into ZMiscData " \
                "(ObjId, OClass, MId, Value_int, Value_string) " \
                "values ('%s', '%s', '%s', '%s', '%s')" % \
                (obj_id, obj_class, misc_id, iv, sv)
                        
            Self.execute (cur, query)

        Self.commit(cur)
        
    def get_misc_data (Self, oclass, oid, mid, check = 0):
        util.log_debug ("get_misc_data()")
        cur = Self.cursor ()
        query = "select Value_int, Value_string from ZMiscData where " \
            "ObjId = '%s' and OClass = '%s' and MId = '%s'" % \
            (oid, oclass, mid)
        
        Self.execute (cur, query)
        
        data = cur.fetchall ()
        Self.commit (cur)

        if not data: 
            if not check: return cfg.MISC_DATA_EMPTY
            else: return 0
            
        assert (len (data) == 1)
        r = data [0]
        
        if check: return 1
        
        return r
        
    def del_misc_data (Self, oclass, oid, mid):
        util.log_debug ("del_misc_data()")
        cur = Self.cursor ()
        query = "delete from ZMiscData where " \
            "ObjId = '%s' and OClass = '%s' and MId = '%s'" % \
            (oid, oclass, mid)
        
        Self.execute (cur, query)

        Self.commit(cur)
        
    def del_misc_data_class (Self, oclass, oid):
        util.log_debug ("del_misc_data_class()")
        cur = Self.cursor ()
        query = "delete from ZMiscData where ObjId = '%s' and OClass = '%s'" % \
            (oid, oclass)
        
        Self.execute (cur, query)

        Self.commit(cur)
        
    #--------------------------------------------
    # Session logging
    
    def log_sql_simple (Self, realm, sid, event):
        util.log_debug ("log_sql_simple()")
        cur = Self.cursor ()
        nextval = Self.__get_nextval(cur, 'ZLog_seq')
        if Self.dbtype_sql == 'SQLite':
            # Setup timestamp from time.localtime()
            import time
            t = time.localtime()
            query = "insert into ZLog (Id, Realm, Id_Session, EventName, EventTime) " \
                "values (%i, '%s', '%s', '%s', '%04i-%02i-%02i %02i:%02i:%02i')" % \
                (nextval, realm, sid, event, t[0], t[1], t[2], t[3], t[4], t[5])
        else:
            query = "insert into ZLog (Id, Realm, Id_Session, EventName) " \
                "values (%i, '%s', '%s', '%s')" % \
                (nextval, realm, sid, event)

        Self.execute (cur, query)

        Self.commit(cur)
        
    def log_sql (Self, realm, sid, event, e_value, e_value2, o_type, 
                                                            o_size, r_addr):
        util.log_debug ("log_sql()")
        cur = Self.cursor ()
        nextval = Self.__get_nextval(cur, 'ZLog_seq')
        if Self.dbtype_sql == 'SQLite':
            # Setup timestamp from time.localtime()
            import time
            t = time.localtime()
            query = "insert into ZLog (Id, Realm, Id_Session, EventName, " \
                "EValue, EValue2, OType, OSize, RAddr, EventTime) values " \
                "(%i, '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%04i-%02i-%02i %02i:%02i:%02i')" % \
                (nextval, realm, sid, event, e_value, e_value2, o_type, \
                o_size, r_addr, t[0], t[1], t[2], t[3], t[4], t[5])
        else:
            query = "insert into ZLog (Id, Realm, Id_Session, EventName, " \
                "EValue, EValue2, OType, OSize, RAddr) values " \
                "(%i, '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s')" % \
                (nextval, realm, sid, event, e_value, e_value2, o_type, \
                o_size, r_addr)
        
        Self.execute (cur, query)
        
        Self.commit (cur)
        
